/*
 * SFO Reader
 * Copyright (C) 2011 Matthieu Milan
 */

#include "file.h"

struct File_impl {
	bool _ioErr;
	File_impl() : _ioErr(false) {}
	virtual ~File_impl() {}
	virtual bool open(const char *path, const char *mode) = 0;
	virtual void close() = 0;
	virtual void seek(uint32_t off) = 0;
	virtual void read(void *ptr, uint32_t size) = 0;
};

struct stdFile : File_impl {
	FILE *_fp;
	stdFile() : _fp(0) {}
	bool open(const char *path, const char *mode) {
		_ioErr = false;
		_fp = fopen(path, mode);
		return (_fp != 0);
	}
	void close() {
		if (_fp) {
			fclose(_fp);
			_fp = 0;
		}
	}
	void seek(uint32_t off) {
		if (_fp) {
			fseek(_fp, off, SEEK_SET);
		}
	}
	void read(void *ptr, uint32_t size) {
		if (_fp) {
			uint32_t r = fread(ptr, 1, size, _fp);
			if (r != size)
				_ioErr = true;
		}
	}
};

File::File() {
	_impl = new stdFile;
}

File::~File() {
	if (_impl) {
		_impl->close();
		delete _impl;
	}
}

bool File::open(const char *filename, const char *directory, const char *mode) {
	if (_impl) {
		_impl->close();
		delete _impl;
		_impl = 0;
	}

	if (!_impl)
		_impl = new stdFile;

	char buf[512];
	snprintf(buf, sizeof(buf), "%s/%s", directory, filename);

	return _impl->open(buf, mode);
}

void File::close() {
	if (_impl) {
		_impl->close();
	}
}

bool File::ioErr() const {
	return _impl->_ioErr;
}

void File::seek(uint32_t off) {
	_impl->seek(off);
}

void File::read(void *ptr, uint32_t size) {
	_impl->read(ptr, size);
}

uint32_t File::getInt(uint32_t offset) {
	seek(offset);

	uint8_t *bytes = (uint8_t *)malloc(4);
	read(bytes, 4);

	uint32_t res = READ_LE_UINT32(bytes);
	free(bytes);

	return res;
}

uint8_t File::getByte(uint32_t offset) {
	seek(offset);
	uint8_t b;
	read(&b, 1);

	return b;
}

uint32_t File::getString(void *ptr, uint32_t offset) {
	uint8_t b = 1;
	uint32_t count = 0;

	while (b != 0) {
		b = getByte(offset + count);
		((uint8_t *)ptr)[count++] = b; // FIXME: peut faire mieux ?
	}

	return count - 1;
}

uint32_t File::getOffsetKeyName(uint32_t offset) {
	return (getByte(21 + (offset << 4)) << 8) | getByte(20 + (offset << 4));
}

uint32_t File::getKey(void *ptr, uint32_t offset) {
	return getString(ptr, getInt(8) + getOffsetKeyName(offset));
}

uint32_t File::getDataType(uint32_t offset) {
	return getByte(23 + (offset << 4));
}

uint32_t File::getSizeOfValue(uint32_t offset) {
	return getInt(24 + (offset << 4));
}

uint32_t File::getValueText(void *ptr, uint32_t offset) {
	return getValueBin(ptr, offset);
}

uint32_t File::getValueBin(void *ptr, uint32_t offset) {
	int sov = getSizeOfValue(offset);
	seek(getInt(12) + getInt(32 + (offset << 4)));
	read(ptr, sov);

	return sov;
}

uint32_t File::getValueInt(uint32_t offset) {
	return getInt(getInt(12) + getInt(32 + (offset << 4)));
}
